import subprocess as sp

import httpx

from invisible_hand.errors import (
    ERR_GIT_CREDENTIAL_HELPER_NOT_FOUND,
    ERR_INVALID_GITHUB_TOKEN,
)


def ensure_git_cached():
    """Would raise GIT_CREDENTIAL_HELPER_NOT_FOUND if git credential not being cached/stored. """
    # detect credential.helper
    procs = sp.Popen(["git", "config", "credential.helper"], stdout=sp.PIPE)
    raw_data = procs.stdout.readline()
    data = raw_data.decode("utf-8").strip()
    if len(data) == 0:
        raise ERR_GIT_CREDENTIAL_HELPER_NOT_FOUND


def ensure_gh_token(token: str):
    """
    Raise token error if token is valid by requesting github
    """
    httpx_headers = httpx.Headers(
        {"User-Agent": "GitHubClassroomUtils/1.0", "Authorization": f"token {token}",}
    )
    res = httpx.get("https://api.github.com/user", headers=httpx_headers)
    if res.status_code != 200:
        raise ERR_INVALID_GITHUB_TOKEN(token=token)



import click
import trio
from tqdm import tqdm

from ..config.github import config_github, config_grant_read_access
from ..ensures import ensure_gh_token
from ..utils.github_entities import Team
from ..utils.github_scanner import query_matching_repos

# Use a "team slug"
# for example: "2019 Teaching-team" -> "2019_teaching-team"


@click.command()
@click.argument("hw_title")
@click.option(
    "--token", default=config_github["personal_access_token"], help="github access token"
)
@click.option("--org", default=config_github["organization"], show_default=True)
@click.option(
    "--team", default=config_grant_read_access["reader_team_slug"], show_default=True
)
def grant_read_access(hw_title, token, org, team):
    """Grant read access right of TA's group to students' homework repo"""
    print("start script")

    ensure_gh_token(token)

    teaching_team = Team(org, team, token)

    repos = query_matching_repos(
        org, github_repo_prefix=f"{hw_title}-", github_token=token, verbose=True
    )
    repo_names = [r["name"] for r in repos]

    # show repos to operate on
    ncols = 3
    cols = [repo_names[i::ncols] for i in range(ncols)]
    print("repos to operate on:")
    for a, b, c in zip(*cols):
        print(f"  {a:<30}{b:<30}{c:<}")

    builder = pbar_builder()
    builder.set_config(total=len(repo_names))
    with builder.build(desc="fired") as fired, builder.build(desc="returned") as returned:

        async def subscribe_to_repo(team: Team, repo_name: str):
            res = await team.add_team_repository_async(repo_name, permission="pull")
            if res.status_code != 204:
                print(res)
            returned.update(1)

        async def async_github():
            async with trio.open_nursery() as nursery:
                for repo_name in repo_names:
                    nursery.start_soon(subscribe_to_repo, teaching_team, repo_name)
                    fired.update(1)
                    fired.refresh()
                return

        trio.run(async_github)


class pbar_builder:
    def __cover_dict(self, src, dst):
        for k, v in src.items():
            dst[k] = v

    def __init__(self):
        self.config = {}
        self.next_po = 0

    def set_config(self, **kwargs):
        self.__cover_dict(kwargs, self.config)

    def build(self, **kwargs):
        cfg = {}
        self.__cover_dict(self.config, cfg)
        self.__cover_dict(kwargs, cfg)
        cfg["position"] = self.next_po
        self.next_po += 1
        return tqdm(**cfg)


if __name__ == "__main__":
    grant_read_access()



import click

from .scripts.add_students import add_students
from .scripts.announce_grade import announce_grade
from .scripts.crawl_classroom import crawl_classroom
from .scripts.event_times import event_times
from .scripts.grant_read_access import grant_read_access
from .scripts.patch_project import patch_project


@click.group(
    invoke_without_command=True, context_settings=dict(help_option_names=["-h", "--help"])
)
@click.pass_context
def cli(ctx):
    """Toolkits for github courses"""
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())


cli.add_command(add_students)
cli.add_command(event_times)
cli.add_command(patch_project)
cli.add_command(announce_grade)
cli.add_command(crawl_classroom)
cli.add_command(grant_read_access)


def main():
    cli()



# Initialize google spreadsheet constants
from configparser import ConfigParser
from pathlib import Path

from .config_util import init_config_file

config_gsheet = None


def init_constants():
    """initialize all constants"""
    global config_gsheet
    template_name = "google_sheet.ini"
    cfile = "gsheet_config.ini"

    if not Path(cfile).exists():
        init_config_file(template_name, cfile)

    config = ConfigParser()
    config.read(cfile)

    # export global variables
    config_gsheet = dict(config["google_spreadsheet"])



import importlib.resources as pkg_resources
from pathlib import Path
from shutil import copyfile

from . import templates


def init_config_file(file_name: str, output_name: str):
    """Copy the file inside ./template to working dir, will fail if file already exists"""
    with pkg_resources.path(templates, file_name) as tmpl_path:
        if Path(output_name).exists():
            print(f"config file `{output_name}` already exists.")
            return
        copyfile(src=tmpl_path, dst=output_name)
        print(f"Create config file: {output_name}")



import atexit
import errno
import os
import shutil
import sys
from pathlib import Path

import click
from bs4 import BeautifulSoup
from halo import Halo
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from ..config.github import config_crawl_classroom

spinner = Halo(stream=sys.stderr)


class Condition_logined:
    def __init__(self):
        pass

    def __call__(self, driver):
        try:
            # will throw an exception if target not found
            _ = driver.find_element_by_id("login_field")
            # print("I can still see login field")
            return False
        except:
            # print("no more login field found")
            return True


# TODO:
#  - error handling
#     - deadline is not due
#
#  - show progress
#     - how many repos in the page that are not loaded
#  - show progress bar in each page
#
#  - report statistics
#     - how many repos
#     - how many students have submitted their hw
#     - how many students didn't submitted their hw and their info


BaseURL = "https://classroom.github.com"


@click.command()
@click.argument("hw_title")
@click.argument("output")
@click.option("--login", default=config_crawl_classroom["login"], show_default=True)
@click.option("--passwd")
@click.option(
    "--classroom_id", default=config_crawl_classroom["classroom_id"], show_default=True
)
def crawl_classroom(hw_title, login, passwd, classroom_id, output):
    """Get student homework submitted version from github classroom
    @output :filename to output your result
    """
    spinner.start()
    if output is not None:
        if Path(output).exists():
            spinner.info(f"Overwrite existing file: {output}")

    spinner.text = "Initializing Chrome driver"
    # We can only use headless=False
    # Because I use my chrome extension helping my web crawling
    # but Google doesn't want to support Chrome extensions in headless mode
    # See https://bugs.chromium.org/p/chromium/issues/detail?id=706008#c5 for more information
    driver = createDriver(headless=False)
    atexit.register(lambda: driver.quit)

    # gh classroom assigment page
    spinner.text = "Opening GitHub Classroom"
    driver.get(BaseURL + "/classrooms/{}/assignments/{}".format(classroom_id, hw_title))

    # user id
    input_id = driver.find_element_by_id("login_field")
    input_id.send_keys(login)

    # user password
    input_passwd = driver.find_element_by_id("password")
    spinner.text = "Login..."

    if passwd is not None:
        input_passwd.send_keys(passwd)
        driver.find_element_by_name("commit").click()
        try:
            _ = WebDriverWait(driver, 0).until(Condition_logined())
            spinner.text = "Login to GitHub Classroom"
            spinner.succeed()
        except:
            spinner.fail("Invalid password")
            sys.exit(1)

    try:
        logined = Condition_logined()
        _ = WebDriverWait(driver, 20).until(logined)
    except:
        spinner.fail("Browser timed out")
        sys.exit(1)

    # check if error
    submitted_infos = []
    not_submitted_list = []
    page_loaded = 0
    while True:
        spinner.text = "Loading page {}".format(page_loaded + 1)
        spinner.start()
        wait_for_page_load(driver)
        page_loaded += 1
        spinner.succeed("Page {} loaded".format(page_loaded))
        soup = BeautifulSoup(driver.page_source, "lxml")
        next_btn = soup.find("a", "next")

        li = soup.find_all("div", "assignment-repo-list-item")
        for i in li:
            user_id = i.find("a", "assignment-repo-github-url").find("h3").contents[0]
            links_to_commit = i.find("a", attrs={"aria-label": "View Submission"})

            if links_to_commit is not None:
                commit_url = links_to_commit["href"]
                commit_hash = commit_url.rsplit("/", 1)[-1]
                submitted_str = "{}-{}:{}".format(hw_title, user_id, commit_hash[0:7])
                submitted_infos.append(submitted_str)
                # print(submitted_str)
            else:
                not_submitted_list.append(f"{hw_title}-{user_id}")

        if next_btn == None:
            break
        driver.get(BaseURL + next_btn["href"])
    spinner.succeed("Success")

    ostream = sys.stdout
    if output is not None:
        ostream = open(output, "w")
        atexit.register(lambda: ostream.close)

    for i in submitted_infos:
        print(i, end=" ", file=ostream)
    print("", file=ostream)

    if len(not_submitted_list) != 0:
        print("Students not submitted:", file=sys.stderr)
        print(not_submitted_list, file=sys.stderr)


def createDriver(headless=True) -> webdriver:

    options = Options()
    options.headless = headless
    options.add_argument("--window-size=1920,1200")
    options.add_extension("./chrome-ghclassroom-waiter.crx")

    cdriver_path = shutil.which("chromedriver")
    if cdriver_path is None:
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), "chromedriver")
    driver = webdriver.Chrome(options=options, executable_path=cdriver_path)
    return driver


def wait_for_page_load(driver):
    try:
        _ = WebDriverWait(driver, 30).until(
            # This id is generated by my chrome extension
            EC.presence_of_element_located(
                (By.ID, "chrome-extension-classroom-waiter-loaded")
            )
        )
    except:
        driver.close()
        raise ("Timeout: cannot load assignment page")


if __name__ == "__main__":
    crawl_classroom()



import click
import requests
from halo import Halo

from ..config.github import config_add_students, config_github
from ..core.color_text import normal, warn
from ..ensures import ensure_gh_token
from ..utils.github_entities import Team
from ..utils.github_scanner import github_headers


def print_table(data, cols=5, wide=15, indent=2):
    """Prints formatted data on columns of given width."""
    n, r = divmod(len(data), cols)
    pat = "{{:{}}}".format(wide)
    line = "\n{}".format(" " * indent).join(pat * cols for _ in range(n))

    # indent the first line
    line = " " * indent + line

    last_line = " " * indent + pat * r
    print(line.format(*data))
    print(last_line.format(*data[n * cols :]))


# Use a "team slug"
# for example: "2019 Students-hello" -> "2019_students-hello"


@click.command()
@click.argument("student_handles", nargs=-1)
@click.option(
    "--token", default=config_github["personal_access_token"], help="github access token"
)
@click.option("--org", default=config_github["organization"], show_default=True)
@click.option(
    "--team", default=config_add_students["default_team_slug"], show_default=True
)
def add_students(student_handles, token, org, team):
    """
    student_handles: github user to add (usernames)
    """
    if len(student_handles) == 0:
        print("required handles")
        return 1

    github_students = student_handles
    github_organization = org
    github_team = team
    github_token = token

    ensure_gh_token(github_token)

    # print('org: {}'.format(github_organization))
    # print('token: {}'.format(github_token))
    # print("students:{}".format(github_students))

    with Halo() as spinner:
        spinner.text = "fetch existing team members from GitHub"
        team = Team(github_organization, team_slug=github_team, github_token=github_token)
        spinner.succeed()

        num_member = len(team.members.keys())
        words = (
            normal.txt("target team: ")
            .kw(f"{github_team}")
            .txt("( ")
            .kw2(num_member)
            .txt(" members) ")
        )
        spinner.info(words.to_str())

    existed_members = set(team.members.keys())
    outside_users = list(set(github_students) - existed_members)

    # print("Users to invite:")
    # print_table(outside_users, cols=5, wide=15)

    invalid_id = []
    with Halo() as spinner:
        spinner.text = ""
        # spinner.start()
        total = len(outside_users)
        for idx, u in enumerate(outside_users, start=1):
            spinner.text = "{}/{} Check valid GitHub username : {}".format(idx, total, u)
            if check_is_github_user(u, github_token) == False:
                invalid_id.append(u)
        spinner.text = "{}/{} Check valid GitHub username".format(total, total)
        spinner.succeed()

    if len(invalid_id) != 0:
        print("Find non-existed github user names:")
        # control strings take space
        print_table([warn.txt("i").to_str() for i in invalid_id], cols=5, wide=25)

    non_member_valid_users = list(set(outside_users) - set(invalid_id))

    # membership info
    membership_infos = {key: "unknown" for key in non_member_valid_users}
    with Halo() as spinner:
        spinner.text = "Check Membership information"
        spinner.start()
        total = len(non_member_valid_users)
        for idx, username in enumerate(non_member_valid_users, start=1):
            spinner.text = "{}/{} Check Membership information : {}".format(
                idx, total, username
            )
            res = team.get_memberships(username)
            if res.status_code == 200:
                membership_infos[username] = res.json()["state"]
        spinner.text = "{}/{} Check Membership information".format(total, total)
        spinner.succeed()

    pending_users = [
        u for u in membership_infos.keys() if membership_infos[u] == "pending"
    ]
    no_memship_users = [
        u for u in membership_infos.keys() if membership_infos[u] == "unknown"
    ]

    print("Users already in pending state (total:{}):".format(len(pending_users)))
    print_table(pending_users)

    print("Users to add (total: {})".format(len(no_memship_users)))
    print_table(no_memship_users)

    failed_users = []
    with Halo() as spinner:
        for user_name in no_memship_users:
            spinner.text = normal.txt("adding user: ").kw(user_name).to_str()
            spinner.start()
            res = team.add_user_to_team(user_name)
            if res.status_code == 200:
                spinner.succeed()
            else:
                failed_users.append(user_name)
                spinner.text += (
                    warn.txt(", return code:").kw(f" {res.status_code} ").to_str()
                )
                spinner.fail()
    failed_users = list(set(failed_users))

    if len(failed_users) != 0:
        print("Users failed to add")
        print_table(failed_users)

    with Halo() as spinner:
        spinner.text = "Adding students successfully"
        spinner.info()


def check_is_github_user(github_id, github_token) -> bool:
    res = requests.get(
        "https://api.github.com/users/{}".format(github_id),
        headers=github_headers(github_token),
    )
    if res.status_code == 200:
        return True
    else:
        return False


if __name__ == "__main__":
    add_students()



# import click

from configparser import ConfigParser
from pathlib import Path

from .config_util import init_config_file

# These variables will be set in init_constants()
config_github = None
config_add_students = None
config_event_times = None
config_announce_grade = None
config_crawl_classroom = None
config_grant_read_access = None

# will be call inside invisible_hand/__init__.py


def init_constants():
    """initialize all constants"""
    global config_github
    global config_announce_grade
    global config_add_students
    global config_event_times
    global config_crawl_classroom
    global config_grant_read_access

    template_name = "github.ini"
    cfile = "github_config.ini"

    if not Path(cfile).exists():
        init_config_file(template_name, cfile)
        # print('propagate config files')
        # sys.exit(1)
    config = ConfigParser()
    config.read(cfile)

    # export global variables
    config_github = dict(config["github"])
    config_event_times = dict(config["event_times"])
    config_add_students = dict(config["add_students"])
    config_announce_grade = dict(config["announce_grade"])
    config_crawl_classroom = dict(config["crawl_classroom"])
    config_grant_read_access = dict(config["grant_read_access"])



from colorama import Back, Fore, Style


def Colored(style):
    return ColoredText(style=style)


class ColoredText:
    def __init__(self, style=None, words=None):
        self.words = words.copy() if words else []
        self.style = style

    def __reset(self):
        return f"{Style.RESET_ALL}{Fore.RESET}{Back.RESET}"

    def __append_word(self, style, s):
        words = self.words.copy()
        words.append(f"{style}{s}{self.__reset()}")
        return ColoredText(style=self.style, words=words)

    def __get_style(self, method: str):
        return getattr(self.style, method)()

    # Define functions here to provide code completion hint
    def txt(self, s):
        return self.__append_word(self.__get_style("txt"), str(s))

    def kw(self, s):
        return self.__append_word(self.__get_style("kw"), str(s))

    def kw2(self, s):
        return self.__append_word(self.__get_style("kw2"), str(s))

    def newline(self):
        words = self.words.copy()
        words.append("\n")
        return ColoredText(style=self.style, words=words)

    def __to_str(self):
        words = []
        for w in self.words:
            if w[0] == "\n":
                pass
            if w[-1] == "\n":
                pass

        words = [w.replace("\n", f"{Style.RESET_ALL}\n") for w in self.words]
        return "".join(words)

    def to_str(self):
        return self.__to_str()

    def __repr__(self):
        return self.__to_str()

    def __str__(self):
        return self.__to_str()

    def strip(self):
        return self.__to_str().strip()

    def __radd__(self, left):
        # convert L.H.S. into a string
        words = [str(left)] + self.words.copy()
        return ColoredText(style=self.style, words=words)

    def __add__(self, other):
        if isinstance(other, ColoredText):
            words = self.words.copy() + other.words.copy()
            return ColoredText(style=other.style, words=words)
        words = self.words.copy() + [str(other)]
        return ColoredText(style=self.style, words=words)



from .cli import cli

# def test():
# g = Gsheet(spreadsheetId='1WlbKDBdhC03m_mFikWwwveDvSQKwKKo20SO9KlbeSqQ')
# g.playground()

# g = Gstudents()

# print( g.gsheet.sheets().get(sid) )
# print(gsheet.get_values())
# print(gsheet.get_values(A1=SAMPLE_RANGE_NAME))
# print(f'title list: {gsheet.get_title_list()}')

if __name__ == "__main__":
    cli()



# github_scanner.py
# Dan Wallach <dwallach@rice.edu>
# Available subject to the Apache 2.0 License
# https://www.apache.org/licenses/LICENSE-2.0

# This file has all the shared logic for interacting with the GitHub service, including keeping
# a local cache to speed things up.

import json
import os
import re
import sys
from datetime import datetime, timezone
from typing import List

import iso8601
import requests
import trio
from requests.models import Response

scanner_cache = {}


def dict_to_pretty_json(d: dict) -> str:
    return json.dumps(d, sort_keys=True, indent=2)


def load_cache(github_organization: str, verbose: bool = True):
    if github_organization in scanner_cache:
        return

    cache_name = ".github-classroom-utils." + github_organization + ".json"
    try:
        if os.path.isfile(cache_name) and os.access(cache_name, os.R_OK):
            with open(cache_name, "r") as file:
                data = json.loads(file.read())
                scanner_cache[github_organization] = data
                if verbose:
                    print("Restored cache: " + github_organization)

    except Exception as e:
        print("Unexpected error loading cache: " + cache_name)
        print(e)
        exit(1)


def store_cache(github_organization: str, verbose: bool = True):
    if github_organization not in scanner_cache:
        if verbose:
            print(
                "Warning: nothing in cache for "
                + github_organization
                + ". Nothing to persist to disk."
            )
        return

    cache_name = ".github-classroom-utils." + github_organization + ".json"
    try:
        with open(cache_name, "w") as file:
            # Pretty-printing the results, even though they're larger. Human-legibility may come in handy.
            file.write(dict_to_pretty_json(scanner_cache[github_organization]))

            if verbose:
                print("Wrote cache for " + github_organization)

    except Exception as e:
        if verbose:
            print("Unexpected error writing cache: " + cache_name)
            print(e)


def github_headers(github_token: str) -> dict:
    """
    Given a GitHub access token, produces a Python dict suitable for passing to requests' headers field.
    If github_token is an empty string, this prints an error message and crashes the program.
    """
    if github_token == "":
        print(
            "\nError: github_token isn't defined, use the --token argument or edit github_config.py to set it"
        )
        exit(1)

    return {
        "User-Agent": "GitHubClassroomUtils/1.0",
        "Authorization": "token " + github_token,
        # needed for the check-suites request
        "Accept": "application/vnd.github.antiope-preview+json",
    }


def fail_on_github_errors(response: Response):
    if response.status_code != 200:
        print("\nRequest failed, status code: %d" % response.status_code)
        print("Headers: %s\n" % dict_to_pretty_json(dict(response.headers)))
        print("Body: %s\n" % dict_to_pretty_json(response.json()))
        raise (
            Exception(
                "Request github error: status code -{}".format(response.status_code)
            )
        )


def query_repos_cached(
    github_organization: str, github_token: str, verbose: bool = True
) -> List[dict]:
    load_cache(github_organization, verbose)

    # How we can tell if our cache is valid: we do a HEAD request to GitHub, which doesn't consume any
    # of our API limit. The result will include an ETag header, which is just an opaque string. Assuming
    # this string is the same as it was last time, then we'll reuse our cached data. If it's different,
    # then something changed, so we'll rescan everything.

    # Ideally, we'd instead use the GitHub v4 GraphQL APIs, which are much, much more efficient than
    # the v3 REST API we're using, but unfortunately, we found some really nasty bugs in the v4
    # GraphQL API (wherein *sometimes* the results we get back are missing entries). This same
    # bug also applies to the v3 "search" endpoint. So fine, we need to page through all of the
    # repos, but at least we can avoid doing it multiple times if nothing's changed.

    # More on how ETag works:
    # https://gist.github.com/6a68/4971859
    # https://developer.github.com/v3/#conditional-requests

    request_headers = github_headers(github_token)

    previous_etag = (
        scanner_cache[github_organization]["ETag"]
        if github_organization in scanner_cache
        else ""
    )
    head_status = requests.head(
        "https://api.github.com/orgs/" + github_organization + "/repos",
        headers=request_headers,
    )
    fail_on_github_errors(head_status)

    current_etag = head_status.headers["ETag"]

    if previous_etag == current_etag:
        if verbose:
            print("Cached result for " + github_organization + " is current")
        return scanner_cache[github_organization]["Contents"]
    else:
        if verbose:
            print("Cached result for " + github_organization + " is missing or outdated")

    if verbose:
        sys.stdout.write("Getting repo list from GitHub")

    all_repos_list = get_github_endpoint_paged_list(
        "orgs/" + github_organization + "/repos", github_token, verbose
    )

    # force it to exist before we create sub-keys
    scanner_cache[github_organization] = {}
    scanner_cache[github_organization]["ETag"] = current_etag
    scanner_cache[github_organization]["Contents"] = all_repos_list

    num_repos = len(all_repos_list)
    if verbose:
        print("Found %d repos in %s" % (num_repos, github_organization))

    if num_repos != 0:
        # if we got an empty list, then something went wrong so don't write it to the cache
        store_cache(github_organization, verbose)

    return all_repos_list


def query_matching_repos(
    github_organization: str,
    github_repo_prefix: str,
    github_token: str,
    verbose: bool = True,
) -> List[dict]:
    """
    This is the function we expect most of our GitHub Classroom utilities to use. Every GitHub repository has
    a URL of the form https://github.com/Organization/Repository/contents, so the arguments given specify
    which organization is being queried and a string prefix for the repositories being matched. The results
    will be cached to dot-files in the current working directory, such that subsequent queries will run
    more quickly, assuming that there are no new repositories in the given organization.

    The results of this call are a list of Python dict objects. The fields that you might find useful
    include:

    clone_url: https link to the repository
        (e.g., 'https://github.com/RiceComp215/comp215-week01-intro-2017-dwallach.git')

    ssh_url: git@github.com link to the repository
        (e.g., 'git@github.com:RiceComp215/comp215-week01-intro-2017-dwallach.git')

    name: the name of the repo itself (e.g., 'comp215-week01-intro-2017-dwallach')

    full_name: the organization and repo (e.g., 'RiceComp215/comp215-week01-intro-2017-dwallach')

    :param github_organization: GitHub Organization being queried.
    :param github_repo_prefix: String prefix to match GitHub Repositories.
    :param github_token: Token for the GitHub API.
    :param verbose: Specifies whether anything should be printed to show the user status updates.
    :return: A list of Python dicts containing the results of the query.
    """
    result = query_repos_cached(github_organization, github_token, verbose)
    return [x for x in result if x["name"].startswith(github_repo_prefix)]


def make_repo_private(repo: dict, github_token: str):
    requests.patch(
        "https://api.github.com/repos/" + repo["full_name"],
        headers=github_headers(github_token),
        json={"private": True},
    )


def get_github_endpoint(endpoint: str, github_token: str, verbose: bool = True) -> dict:
    result = requests.get(
        "https://api.github.com/" + endpoint, headers=github_headers(github_token)
    )
    fail_on_github_errors(result)

    return result.json()


def get_github_endpoint_paged_list(
    endpoint: str, github_token: str, verbose: bool = True, **kwarg
) -> List[dict]:
    page_number = 1
    result_list = []

    while True:
        if verbose:
            sys.stdout.write(".")
            sys.stdout.flush()

        headers = github_headers(github_token)
        result = requests.get(
            "https://api.github.com/" + endpoint,
            headers=headers,
            params={"page": page_number, **kwarg} if page_number > 1 else {**kwarg},
        )
        fail_on_github_errors(result)
        page_number = page_number + 1

        result_l = result.json()

        if len(result_l) == 0:
            if verbose:
                print(" Done.")
            break

        result_list = result_list + result_l

    return result_list


async def get_github_endpoint_paged_list_async(
    client, endpoint: str, **kwarg
) -> List[dict]:
    paged_res = {}

    async def get_page_res(url, page, **kwarg):
        return await client.get(url, params={"page": page, **kwarg})

    # first query, get total number of pages
    page_number = 1
    # res_first_page = client.get('https://api.github.com/' + endpoint, params={'page':page_number, **kwarg})
    res_first_page = await get_page_res(
        f"https://api.github.com/{endpoint}", page_number, **kwarg
    )
    paged_res[page_number] = res_first_page.json()

    # serch through media string and find the last page index
    media_link_str = str(res_first_page.headers.get("link"))
    # print(res_first_page.headers)
    link_to_last_page = re.compile('<https:[^>]*?\?page=(\d*)[^>]*>; rel="(last)"')
    match = link_to_last_page.search(media_link_str)
    last_page_idx = 1
    if match != None:
        last_page_idx = match.group(1)

    async with trio.open_nursery() as nursery:

        async def collect_res(i):
            res = await get_page_res(
                f"https://api.github.com/{endpoint}", page_number, **kwarg
            )
            paged_res[i] = res.json()

        for i in range(1, last_page_idx + 1):
            nursery.start_soon(collect_res, i)

    result_list = []
    for i, res in paged_res.items():
        # print(f'gather page:{i}')
        result_list = result_list + res
    return result_list


# And now for a bunch of code to handle times and timezones. This is probably going to
# require Python 3.7 or later.

LOCAL_TIMEZONE = datetime.now(timezone.utc).astimezone().tzinfo


def datetime_to_local_timezone(dt: datetime) -> str:
    return localtime_from_timestamp(dt.timestamp())


def localtime_from_timestamp(timestamp: float) -> str:
    return str(datetime.fromtimestamp(timestamp, LOCAL_TIMEZONE))


def localtime_from_iso_datestr(date_str: str) -> str:
    return datetime_to_local_timezone(iso8601.parse_date(date_str))


def student_name_from(github_prefix: str, repo_name: str) -> str:
    """
    Given a GitHub repo "name" (e.g., "comp215-week01-intro-danwallach") return the username suffix at the
    end ("danwallach"). If it's not there, the result is an empty string ("").
    """
    m = re.search(github_prefix + "-(.*)$", repo_name)
    if not m:
        return ""  # something funny in the name, so therefore not matching
    else:
        # there might be a trailing dash and digits if the student did the clone thing multiple times
        # also, we're converting everything to lower-case
        return re.sub("-\\d+$", "", m.group(1)).lower()


def desired_user(github_prefix: str, ignore_list: List[str], name: str) -> bool:
    """
    Given a GitHub repo "name" (e.g., "comp215-week01-intro-2017-danwallach"), returns true or false if that
    project is something we're trying to grade now, based on the specified prefix as well as the list of graders
    (to be ignored). Since we might be dealing with student groups, which can give themselves their own group names,
    this function defaults to True, unless it finds a reason to say False.
    """

    # This is inefficient, since we're not caching the
    # result, but performance doesn't really matter here.
    lower_case_ignore_list = [x.lower() for x in ignore_list]

    m = student_name_from(github_prefix, name).lower()
    return (
        m != ""
        and name.startswith(github_prefix)
        and name != github_prefix
        and m not in lower_case_ignore_list
    )



import sys

import httpx
import requests
from requests.models import Response

from ..config.github import config_github
from ..core.color_text import warn
from .github_scanner import (
    get_github_endpoint,
    get_github_endpoint_paged_list,
    github_headers,
)


def main():
    print("nothin to run")


class Team:
    def __init__(self, org, team_slug="", github_token=""):
        self.info_sucess = False
        self.id = ""
        self.org = org
        self.team_slug = team_slug
        self.members = dict()
        self.github_token = github_token
        self._get_members()

        # async
        self.httpx_headers = httpx.Headers(
            {
                "User-Agent": "GitHubClassroomUtils/1.0",
                "Authorization": "token " + github_token,
                # needed for the check-suites request
                "Accept": "application/vnd.github.antiope-preview+json",
            }
        )
        self.async_client = httpx.AsyncClient(headers=self.httpx_headers)

    def add_user_to_team(self, user_name) -> Response:
        res = requests.put(
            "https://api.github.com/teams/{}/memberships/{}".format(self.id, user_name),
            headers=github_headers(self.github_token),
        )
        return res

    def add_team_repository(self, repo, permission="pull") -> Response:
        """If permission=pull, it is equivalent to make this team subscribe to this repo"""
        api_param = {"team_id": self.id, "owner": self.org, "repo": repo}
        return requests.put(
            "https://api.github.com/teams/{team_id}/repos/{owner}/{repo}".format(
                **api_param
            ),
            headers=github_headers(self.github_token),
        )

    async def add_team_repository_async(self, repo, permission="pull") -> Response:
        """If permission=pull, it is equivalent to make this team subscribe to this repo"""
        api_param = {"team_id": self.id, "owner": self.org, "repo": repo}
        # print(f'param: {api_param}')
        return await self.async_client.put(
            "https://api.github.com/teams/{team_id}/repos/{owner}/{repo}".format(
                **api_param
            )
        )

    def get_memberships(self, user_name: str) -> Response:
        return requests.get(
            "https://api.github.com/teams/{}/memberships/{}".format(self.id, user_name),
            headers=github_headers(self.github_token),
        )

    def _require_team_id(func):
        def decorator(self):
            if self.id == "":
                self._get_team_id()
            func(self)

        return decorator

    @_require_team_id
    def _get_members(self):
        try:
            # res = get_github_endpoint(
            #     endpoint="teams/{}/members".format(self.id),
            #     github_token=self.github_token,
            # )
            res = get_github_endpoint_paged_list(
                endpoint="teams/{}/members".format(self.id),
                github_token=self.github_token,
                verbose=False,
            )
        except:
            prompt_cannot_fetch_team(org=self.org, team_slug=self.team_slug)
            sys.exit(1)
        for user in res:
            self.members[user["login"]] = user
        # print(res)

    def _get_team_id(self):
        try:
            result = get_github_endpoint(
                "orgs/{}/teams/{}".format(self.org, self.team_slug),
                config_github["personal_access_token"],
            )
        except:
            prompt_cannot_fetch_team(org=self.org, team_slug=self.team_slug)
            sys.exit(1)
        self.id = result["id"]


def prompt_cannot_fetch_team(org: str, team_slug: str):
    print(warn.txt("Cannot fetch team id").to_str)
    print("Team info:")
    print(f"    org: {org}")
    print(f"    team_slug: {team_slug}")


if __name__ == "__main__":
    main()



# github_event_times.py
# Dan Wallach <dwallach@rice.edu>
# Available subject to the Apache 2.0 License
# https://www.apache.org/licenses/LICENSE-2.0

import re
import sys
from datetime import datetime, timedelta
from pathlib import Path
from pprint import pprint
from typing import List, NamedTuple, Tuple

# import argparse
import click
import iso8601
from halo import Halo
from tabulate import tabulate

from ..config.github import config_event_times, config_github
from ..ensures import ensure_gh_token
from ..utils.github_entities import Team
from ..utils.github_scanner import (
    LOCAL_TIMEZONE,
    get_github_endpoint_paged_list,
    localtime_from_iso_datestr,
)


def is_deadline_passed(dl: datetime, submit: datetime) -> Tuple[bool, timedelta]:
    dl = dl.astimezone(submit.tzinfo)
    if dl < submit:
        return True, submit - dl
    else:
        return False, dl - submit


class commitInfo(NamedTuple):
    commit_hash: str  # 7 characters hash value
    pushed_time: str
    msg: str
    repo: str


class GitHubRepoInfo(NamedTuple):
    name: str
    commit_hash: str  # 7 characters hash value


# python3 github_event_times.py hw0-ianre657:cb75e99


@click.command()
@click.argument("input-file")
@click.option(
    "--token", default=config_github["personal_access_token"], help="github access token"
)
@click.option("--org", default=config_github["organization"], show_default=True)
@click.option(
    "--deadline",
    default=config_event_times["deadline"],
    show_default=True,
    help='deadline format: "yyyy-mm-dd" or any ISO8601 format string (timezone will be set to local timezone)',
)
@click.option("--target-team", default=None, help="specific team to operate on")
def event_times(input_file, org, token, deadline, target_team):
    """
    input-file: file contains list of repo-hash.

    repo-hash : string in <repo>:<hash> format
            hw0-ianre657:cb75e99
    """
    global github_organization
    global github_token

    try:
        parsed_repos = get_repo_infos(input_file)
    except FileNotFoundError as e:
        print(str(e))
        return
    ensure_gh_token(token)
    spinner = Halo(stream=sys.stderr)

    github_organization = org
    github_token = token

    print(f"deadline: {deadline}")

    submit_deadline = iso8601.parse_date(deadline)
    submit_deadline = submit_deadline.replace(tzinfo=LOCAL_TIMEZONE)

    spinner.info(f"Deadline : {submit_deadline}")
    success_group = []
    fail_group = []
    spinner.start("Start to check late submissions")

    # get team membershup info
    if target_team is not None:
        only_team_members = set(
            Team(
                org=github_organization, team_slug=target_team, github_token=github_token
            ).members.keys()
        )

    for idx, repo in enumerate(parsed_repos, start=1):
        # print("get commit time for {}".format(repo))
        if target_team is not None:
            import re

            user_id = re.sub("hw[\d]+-", "", repo.name)
            # print(f'user_id :{user_id}')
            if user_id not in only_team_members:
                continue
        spinner.text = f"({idx}/{len(parsed_repos)}) Checking {repo.name}"
        result = getRepoCommitTime(
            org=github_organization, repo=repo.name, commit_hash=repo.commit_hash
        )
        for r in result:
            # print(r)
            passed, delta = is_deadline_passed(
                submit_deadline, iso8601.parse_date(r.pushed_time)
            )
            if passed:
                fail_group.append(
                    {
                        "repo-name": r.repo,
                        "commit-hash": r.commit_hash,
                        "time-passed": delta,
                        "last-pushtime": r.pushed_time,
                    }
                )
            else:
                success_group.append((r, delta))
                # print(f'{r}: {delta} later')
    spinner.succeed("Check finished")
    print("=" * 20, "REPORT", "=" * 20)
    print(f"Total submissions : {len(parsed_repos)}")
    print(f"late submissions: {len(fail_group)}")
    print(f"Submission Deadline: {submit_deadline}")
    print(tabulate(fail_group, headers="keys"))


def get_repo_infos(filename: str) -> List[GitHubRepoInfo]:
    st = Path(filename).read_text()
    infos = st.split()
    result = []
    for r in infos:
        repo, commit_hash = r.split(":")
        result.append(GitHubRepoInfo(repo, commit_hash))
    return result


def getRepoCommitTime(org: str, repo: str, commit_hash: str) -> List[commitInfo]:
    global github_token
    response = get_github_endpoint_paged_list(
        f"repos/{org}/{repo}/events", github_token, verbose=False
    )
    event_list = [x for x in response if x["type"] == "PushEvent"]
    # find the localtiome of the given commit SHA that is pushed.
    msgs = []
    for event in event_list:
        try:
            github_id = event["actor"]["login"]
            date = localtime_from_iso_datestr(event["created_at"])
            target_commits = [
                x for x in event["payload"]["commits"] if x["sha"][0:7] == commit_hash
            ]

            if len(target_commits) == 0:
                continue
            elif len(target_commits) >= 2:
                print("SHA Conflict for user: {}".format(github_id))
            else:
                target_commit = target_commits[0]

                commit_message = tex_escape(
                    target_commit["message"].splitlines()[0]
                )  # only the first line if multiline
                commit_hash = target_commit["sha"][0:7]

                result = commitInfo(commit_hash, date, commit_message, repo)
                # result_str = "{} sha:{}, msg:{}, date:{}".format(github_id, commit_hash, commit_message, date)
                msgs.append(result)
        except KeyError:
            print("Error: malformed event!")
            pprint(event)
    return msgs


# https://stackoverflow.com/questions/16259923/how-can-i-escape-latex-special-characters-inside-django-templates


def tex_escape(text: str) -> str:
    """
    :param text: a plain text message
    :return: the message escaped to appear correctly in LaTeX
    """
    conv = {
        "&": r"\&",
        "%": r"\%",
        "$": r"\$",
        "#": r"\#",
        "_": r"\_",
        "{": r"\{",
        "}": r"\}",
        "~": r"\textasciitilde{}",
        "^": r"\^{}",
        "\\": r"\textbackslash{}",
        "<": r"\textless{}",
        ">": r"\textgreater{}",
    }
    regex = re.compile(
        "|".join(
            re.escape(str(key))
            for key in sorted(conv.keys(), key=lambda item: -len(item))
        )
    )
    return regex.sub(lambda match: conv[match.group()], text)


if __name__ == "__main__":
    event_times()



from colorama import Back, Fore, Style


class WarnStyle:
    @classmethod
    def txt(cls) -> str:
        return f"{Back.RED}{Fore.BLACK}"

    @classmethod
    def kw(cls) -> str:
        return f"{Back.RED}{Fore.WHITE}{Style.BRIGHT}"

    @classmethod
    def kw2(cls) -> str:
        return f"{Back.RED}{Fore.YELLOW}{Style.BRIGHT}"


class NormalStyle:
    @classmethod
    def txt(cls) -> str:
        return f"{Back.GREEN}{Fore.BLACK}"

    @classmethod
    def kw(cls) -> str:
        return f"{Back.GREEN}{Fore.WHITE}{Style.BRIGHT}"

    @classmethod
    def kw2(cls) -> str:
        return f"{Back.GREEN}{Fore.RED}{Style.NORMAL}"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['invisible_hand',
 'invisible_hand.config',
 'invisible_hand.config.templates',
 'invisible_hand.core',
 'invisible_hand.core.color_text',
 'invisible_hand.scripts',
 'invisible_hand.utils']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4>=4.9.1,<5.0.0',
 'click>=7.1.2,<8.0.0',
 'colorama>=0.4.3,<0.5.0',
 'gitpython>=3.1.7,<4.0.0',
 'google-api-python-client>=1.10.0,<2.0.0',
 'google-auth-httplib2>=0.0.4,<0.0.5',
 'google-auth-oauthlib>=0.4.1,<0.5.0',
 'halo>=0.0.30,<0.0.31',
 'httpx>=0.14.1,<0.15.0',
 'iso8601>=0.1.12,<0.2.0',
 'lxml>=4.5.2,<5.0.0',
 'oauth2client>=4.1.3,<5.0.0',
 'prompt-toolkit>=3.0.6,<4.0.0',
 'pydantic>=1.7,<2.0',
 'pygsheets>=2.0.3,<3.0.0',
 'requests>=2.24.0,<3.0.0',
 'selenium>=3.141.0,<4.0.0',
 'tabulate>=0.8.7,<0.9.0',
 'tqdm>=4.48.2,<5.0.0',
 'trio>=0.16.0,<0.17.0',
 'xlsxwriter>=1.3.3,<2.0.0']

entry_points = \
{'console_scripts': ['hand = invisible_hand.cli:main']}

setup_kwargs = {
    'name': 'invisible-hand',
    'version': '0.2.4',
    'description': 'Automate your workflow with github classroom',
    'long_description': '# Invisible Hand\n\nInvisible Hand is a set of tools to manage your classroom inside github organization. It utilizes `Google Sheets` , `GitHub Classroom` and `GitHub` seamlessly.\n\n## Installation\n\n#### 1. Install via pip\n\n `pip install invisible-hand`\n\n#### 2. Install chromedriver\n\n* OSX\n\n `brew cask install chromedriver`\n* Ubuntu\n\n `apt install chromium-chromedriver`\n\n## Config Files\n\n__Invisible Hand__ operates on 2 config files: `github_config.ini` and `gsheet_config.ini` .\n\nCalling `hand` will automatically propagate these files into working directory with default template. Make sure you have configured them correctly before using them.\n\nAdditionally, if you want to use [ `Announce Grade` ](#announce-grade), follow [here](https://pygsheets.readthedocs.io/en/stable/authorization.html) to get your client secret file and rename it to __client_secret.json__\n\n## Usage\n\n### hand\n\nThe root command\n\n<details>\n<summary><b>Show Detail Information</b></summary>\n\n#### Help Message\n\nappend `-h` or `--help` options get help messages\n\n``` sh\n$ hand -h\nUsage: hand [OPTIONS] COMMAND [ARGS]...\n\n    Toolkits for compiler-f19\n\nOptions:\n    -h, --help         Show this message and exit.\n\nCommands:\n    add-students       student_ids: ids to add\n    announce-grade     announce student grades to each hw repo\n    event-times        repo-hashes : list of <repo>:<hash> strings ex:...\n    grant-team-access  Add students into a github team hw-prefix: prefix for...\n    patch-project      Patch to student homeworks\n```\n\n</details>\n\n---\n\n### Add Students\n\ninvite users into your github organization\n\n<details><summary><b>Show Detail Information</b></summary>\n\n#### Format\n\n``` sh\nhand add-students [github_handle]...\n```\n\n> Use `-h` to see more detailed information about this command.\n\ngithub_handle: github accounts\n\n#### Example\n\n``` sh\nhand add-students ianre657 cmprfk1 cmprfk2 cmprfk3\n```\n\n#### Config file\n\n* github_config.ini\n  + `[github]:personal_access_token`\n  + `[github]:organization`\n  + `[add_students]:default_team_slug`\n\n#### FAQ\n\n* Some students report that they didn\'t get the invitation email.\n\n    Invite student into your organization from their email. This should be Github\'s issue.\n\n    > about 2 of 80 students got this issue from our previous experience.\n\n</details>\n\n---\n\n### Grant specific team read access to H. W. repos\n\nGrant read access right of TA\'s group to students\' homework repo\n\n<details><summary><b>Show Detail Information</b></summary>\n\n#### Config File\n\n* __github_config.ini__:\n  + `[grant_read_access]:reader_team_slug` : team slug of your TA\'s group\n\n#### Format\n\n``` shell\nhand grant-read-access <hw_title>\n```\n\n#### Example\n\n``` shell\nhand grant-read-access hw3\n```\n\n</details>\n\n---\n\n### Patch Project\n\nPatch to student homework repositories.\n\n<details><summary><b>Show Detail Information</b></summary>\n\n#### Config File\n\n* __github_config.ini__:\n  + `[github]:personal_access_token`\n  + `[github]:organization`\n* __gsheet_config.ini__\n  + `[google_spreadsheet]:spreadsheet_url`\n\n#### Format\n\n``` sh\nhand patch-project <hw_title> [--only-repo] <patch_branch>\n```\n\n> Use `-h` to see more detailed information about this command.\n\nBelow is the standard workflow to follow.\n\n#### Workflow\n\ntake homework : __ `hw3` __(the title of your homework in github classroom) for example:\n\n1. The repo __ `tmpl-hw3` __ would be your template for initializing homeworks.\n2. Create another repo to update your template, let\'s say: __ `tmpl-hw3-revise` __\n3. Inside __ `tmpl-hw3-revise` __, create a revision branch __ `1-add-some-new-feature` __ (whatever you like) and an issuse named as the branch name (in this example, __ `1-add-some-new-feature` )__, which will be the content of your PR message.\n4. Open github-classroom, choose your assignment (__ `hw3` __) and disable `assignment invitation URL` of __ `hw3` __.\n5. Create an PR to your template repo (__ `hw3` __) by using this command.\n\n    \n\n``` sh\n    hand patch-project hw3 --only-repo="tmpl-hw3" 1-add-some-new-feature\n    ```\n\n6. Accept the PR in your template repository (__ `tmpl-hw3` __). After that, enable the `assignment invitation URL` of `hw3` in GitHub Classroom. Now you have succcessfully updated your template repo.\n7. Create PRs to students template repositories ( `hw3-<their github id>` ) by running the scirpt as followed.\n\n    \n\n``` sh\n    hand patch-project hw3 1-add-some-new-feature\n    ```\n\n    This script would patch to every repository that uses __hw3__ as the prefix under your GitHub organization.\n\n8. Merge the revision brnach __ `1-add-some-new-feature` __ into `master` in your __ `tmpl-hw3-revise` __ repo. After this step, all documents are updated.\n\n#### Demo (Deprecated)\n\n<img src="./demos/patcher.gif" alt="patcher-demo-video" width="640">\n\n</details>\n\n---\n\n### Crawl Classroom\n\nCrawling homework submission data from Github Classroom\n\n<details><summary><b>Show Detail Information</b></summary>\n\nThis is a web crawler for Github Classroom, which is the input of [ `Event Times` ](#event-times)\n\n#### Config File\n\n* __github_config.ini__:\n  + `[crawl_classroom]:login` : your login id in Github Classroom\n  + `[crawl_classroom]:classroom_id` : the id field of your classroom RESTful page URL. (see the image below)\n\n    <img src="./imgs/clsrm_id.png" alt="id field in the url of github classroom" width="640">\n\n#### Format\n\n``` sh\nhand crawl-classroom [OPTIONS] HW_TITLE OUTPUT\n```\n\n> Use `-h` to see more detailed information about this command\n\n#### Example\n\n``` shell\nhand crawl-classroom --passwd=(cat ~/cred/mypass) hw5 hw5_handle.txt\n```\n\n> This example suppose you use Fish Shell and store your password inside `~/cred/mypass`\nUsers should type their passsword inside the pop-up window if they don\'t provide their password in the argument\n\n#### FAQ\n\n* ChromeDriver\n\n  \n\n``` \n  selenium.common.exceptions.SessionNotCreatedException: Message: session not created: This version of ChromeDriver only supports Chrome version 79\n  ```\n\n  upgrade your chromedriver via `brew cask upgrade chromedriver`\n* All students not submitted\n  + Remember to set deadline of hw on the GitHub classroom (note that deadline can only be set at a future time)\n\n#### Demo\n\n<img src="./demos/github_classroom_craw.gif" alt="github_classroom_craw" width="640">\n\n</details>\n\n---\n\n### Event Times\n\nRetrieve information about late submissions\n\n<details><summary><b>Show Detail Information</b></summary>\n\n#### What it actually does\n\nCompare the last publish-time of specific git commit in each repository and print out which passed the deadline.\n\n#### Config File\n\n* __github_config.ini__:\n  + `[event_times]:deadline` deadline for homework, in ISO8601 compatible format.\n\n    For example `2019-11-12 23:59:59` (the timezone is set to your local timezone as default).\n\n#### Format\n\n``` sh\nhand event-times <input_file> [--deadline="yyyy-mm-dd"]\n```\n\n__input-file__: file contains list of `repo-hash` .\n\n__repo-hash__ : in the format of `<repo>:<git commit hash>` , (for example: hw0-ianre657:cb75e99)\n\nGithub API use the first 7 characters of a commit\'s SHA-1 value to communicate, so the hash we used here is in the length of 7.\n\n> The input pairs `repo:hash` could be retrieve from [ `Crawl Classroom` ](#crawl-classroom).\n\n__ `--deadline` __: it will use the variable inside `github_config.ini` as default.\n\n__ `--target-team` __ (optional): teams to operate on (use team-slug)\n\n#### Example\n\n``` sh\nhand event-times  --target-team="2020-inservice-students" --deadline="2019-11-12 23:59:59"  hw1-handin-0408.txt\n```\n\n#### Demo (need to be updated)\n\n<img src="./demos/event_times.gif" alt="event-times-demo-video" width="640">\n\n</details>\n\n---\n\n### Announce Grade\n\nPublish feedbacks by creating Issue to student\'s homework repo.\n\n<details><summary><b>Show Detail Information</b></summary>\n\n#### Explanation\n\nIn every homework project, we would create a git repository for every student. Take homework `hw3` with two students `Anna` and `Bella` for example, we expect there would be two repos under our github organization, which is `hw3-Anna` and `hw3-Bella` .\nDuring our grading process, T. A.s would record every grade in a google sheet with a tab named `hw3` and a markdown file for each student in every assignment as their feedbacks.\nAfter their homeworks being graded, we use this code to publish student\'s grade by creating `Issue` s named `Grade for hw3` to each of their github repositories.\n\nThe markdown file for feedbacks contains python template strings, and those strings are the column names inside our google sheet tab `hw3` . One template string we used is students grades, this makes managing grades more easily.\n\nTo use this code, you need to fufill some assumptions.\n\nLets say you\'re about to announce the grade for `hw3` :\n\n* prequisite:\n  1. a git repo to store student feedback templates, which strutured as followed:\n\n``` bash\n  . Hw-manager # root of your git repo\n  ├── hw3\n  │\xa0\xa0 └── reports\n  │\xa0\xa0     ├── 0411276.md\n  │\xa0\xa0     ├── 0856039.md\n  │\xa0\xa0     └── 0956323.md\n  └── hw4 # other homework dir\n```\n\nand inside `0411276.md` , it would be:\n\n``` markdown\n  # Information\n\n  + Student Id: ${student_id}\n  + Grade : ${grade}\n\n  # <Some other important things...>\n  ...\n```\n\n  2. a google sheet to store student information\n\n    | student_id | grade |\n    | :--------: | :---: |\n    |  0856039   |  93   |\n    |  0411276   |  80   |\n\n#### Config file\n\n* __github_config.ini__\n  + `[github]:personal_access_token`\n  + `[github]:organization`\n  + `[announce_grade]:feedback_source_repo` (e.x.: Hw-manager)\n* __gsheet_config.ini__\n  + `[google_spreadsheet]:spreadsheet_id`\n* __client_secret.json__ (follow [here](https://pygsheets.readthedocs.io/en/stable/authorization.html) to download your oauth2 secret file and renamed it to __client_secret.json__)\n\n#### instructions to follow\n\n1. Edit config files properly.\n2. Create feedbacks for students in your `feedback_source_repo`\n3. use this script\n\n#### Format\n\n``` sh\nhand announce-grade <hw_title> [--only-id <student_id>]\n```\n\noption:\n\n`--only-id` : only patch to this student id\n\n#### Example\n\n``` sh\nhand announce-grade hw3 --only-id 0411276\n```\n\n</details>\n\n---\n',
    'author': 'Ian Chen',
    'author_email': 'ianre657@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



from pprint import pformat
from typing import Dict, List, Optional

from .core.color_text import warn
from .urls import url_for_issue


class ERR_UNIQUE_STUDENT_ID(Exception):
    def __init__(self, explanation: str, instances: List[Dict]):
        super().__init__(self)
        self.explanation: str = explanation
        self.instances: List[Dict] = instances

    def __str__(self):
        words = [self.explanation, pformat(self.instances)]
        return "\n".join(words)


class ERR_REQUIRE_NO_SPACE(Exception):
    def __init__(self, explanation: str, instances: List[str]):
        super().__init__(self)
        self.explanation: str = explanation
        self.instances: List[str] = instances

    def __str__(self):
        words = [self.explanation, pformat(self.instances)]
        return "\n".join(words)


class ERR_GIT_CREDENTIAL_HELPER_NOT_FOUND(Exception):
    def __init__(self):
        super().__init__(self)

    def _get_git_cred_warning(self):
        def ljust(word, leng):
            return "{0:<{1}}".format(word, leng)

        leng = 65
        words = (
            warn.newline()
            .txt(ljust(" Not detect any credential helper inside git, use: ", leng))
            .newline()
            + warn.kw(
                ljust("  git config --global credential.helper cache", leng)
            ).newline()
            + warn.txt(ljust(" before going any further.", leng)).newline()
            + warn.txt(" Or see: ")
            + warn.kw2(f" {url_for_issue(issue_num=8)}   ").newline()
            + warn.txt(ljust(" for more information", leng))
            + warn.newline()
        )
        return words.to_str()

    def __str__(self):
        return self._get_git_cred_warning()


class ERR_INVALID_GITHUB_TOKEN(Exception):
    def __init__(self, token: Optional[str] = None):
        self.token = token
        info = "Invalud token" + f": {token}" if token else ""
        super().__init__(self, info)

    def __str__(self) -> str:
        if self.token:
            words = (
                warn.newline()
                + warn.txt(" Invalid Github token:").kw(f" {self.token} ").newline()
                + warn.txt(" please update ").kw2("github_config.ini").newline()
            )
        else:
            words = (
                warn.newline()
                + warn.txt(" Invalid Github token ").newline()
                + warn.txt(" please update ").kw2("github_config.ini ").newline()
            )
        return str(words)



from colorama import init as colorama_init

from .config.github import init_constants as github_init_constants
from .config.gsheet import init_constants as gsheet_init_constants

colorama_init(autoreset=True)

github_init_constants()
gsheet_init_constants()



import string
from typing import Dict, List, Optional

import pygsheets
from pydantic import BaseModel
from xlsxwriter.utility import xl_col_to_name

from ..config.gsheet import config_gsheet
from ..errors import ERR_REQUIRE_NO_SPACE, ERR_UNIQUE_STUDENT_ID


class StudentInfo(BaseModel):
    # student_id must be unique
    student_id: str

    github_handle: str
    name: Optional[str] = None
    email: Optional[str] = None


def contains_space(word: str) -> bool:
    """test if there's any white space in the give word"""
    for s in word:
        if s.isspace():
            return True
    return False


class pygsheetInteractor:
    def __init__(self, pyg=pygsheets):
        self.gc = pyg.authorize()
        self.sht = None

    def open_by_url(self, url):
        self.sht = self.gc.open_by_url(url)

    def _get_wks_by_title(self, title):
        if self.sht is None:
            raise RuntimeError("Call open_by_url before using any function")
        return self.sht.worksheet_by_title(title)

    def get_all_record(self, title, head=1) -> List[Dict]:
        wks = self._get_wks_by_title(title)
        endtag = f"{xl_col_to_name(wks.cols-1)}{wks.rows}"

        grange = pygsheets.GridRange(end=endtag)

        data = wks.get_values(
            grange=grange, include_tailing_empty=True, include_tailing_empty_rows=False
        )

        idx = head - 1
        keys = data[idx]
        field_index_map = {
            field: n for n, field in enumerate(keys) if field.strip() != ""
        }
        vals = []
        for row in data[idx + 1 :]:
            val = {field: row[i] for field, i in field_index_map.items()}
            vals.append(val)
        return vals


class Gstudents:
    def __init__(self, url=config_gsheet["spreadsheet_url"], actor=None):
        self.actor = actor if actor is not None else pygsheetInteractor()
        self.actor.open_by_url(url)

        self.config = {
            "main_wks_title": "StudentInfo",
            "main_wks_required_fields": ["github_handle", "student_id", "name", "email",],
            "main_kws_key_field": "student_id",
        }

        # parse and get student infos
        student_infos = self._build_student_info_records()
        self._ensure_no_space_in_student_fields(student_infos)
        self._ensure_unique_student_id(student_infos)
        self.student_infos: List[StudentInfo] = student_infos

    def _build_student_info_records(self) -> List[StudentInfo]:
        """Get student info from google sheet and do preliminary test"""
        records = self.actor.get_all_record(self.config["main_wks_title"])
        student_infos = []

        def is_valid(word: str) -> bool:
            """if the given word contains any printable character.
            """
            word = word.translate({ord(c): None for c in string.whitespace})
            return len(word) > 0

        for record in records:
            sid = record.get("student_id")
            handle = record.get("github_handle")
            name = record.get("name")
            email = record.get("email")
            if is_valid(sid) and is_valid(handle):
                student_infos.append(
                    StudentInfo(
                        student_id=sid, github_handle=handle, name=name, email=email,
                    )
                )
            elif is_valid(sid) or is_valid(handle):
                fields: Dict[str, str] = self.config["main_wks_required_fields"]
                uncomplete_record = {f: record.get(f) for f in fields}
                # @todo: show warning using log
                print(f"skip: uncomplete field - {uncomplete_record}")
        return student_infos

    def get_student_infos(self) -> List[StudentInfo]:
        return self.student_infos

    def get_students(self) -> List[Dict[str, str]]:
        """Return the student infos in dict"""
        result = []
        for s in self.student_infos:
            result.append(dict(s))
        return result

    def get_student_info(self, student_id: str) -> Optional[StudentInfo]:
        """Fetch a single students info, will traverse the entire table
        Use get_students manually if you need to get each student's info
        """
        students: List[StudentInfo] = self.student_infos
        for s in students:
            if student_id == s.student_id:
                return s
        return None

    def get_student(self, student_id: str) -> Optional[Dict[str, str]]:
        student = self.get_student_info(student_id)
        if student:
            return dict(student)
        return None

    def left_join(self, right_sheet_title) -> List[Dict[str, str]]:
        right_dicts = self.actor.get_all_record(right_sheet_title)

        def left_matching(left, right, key_field: str):
            for info in left:
                for r in right:
                    l_val = getattr(info, key_field)
                    r_val = r.get(key_field)
                    if l_val and r_val and l_val == r_val:
                        yield dict(info), r

        joined_infos = []
        for left, right in left_matching(
            self.student_infos, right_dicts, self.config["main_kws_key_field"]
        ):
            student = {**left, **right}
            joined_infos.append(student)
        return joined_infos

    def _ensure_unique_student_id(self, student_infos: List[StudentInfo]):
        ids = set()
        errors: List[Dict] = []
        for info in student_infos:
            if info.student_id not in ids:
                ids.add(info.student_id)
            else:
                errors.append(dict(info))
        if len(errors) > 0:
            raise ERR_UNIQUE_STUDENT_ID(
                explanation="Student ID should be unique, found:", instances=errors
            )

    def _ensure_no_space_in_student_fields(self, student_infos: List[StudentInfo]):
        errors: List[str] = []
        for info in student_infos:
            for field in self.config["main_wks_required_fields"]:
                value = getattr(info, field)
                if contains_space(value):
                    errors.append(value)
        if len(errors) > 0:
            raise ERR_REQUIRE_NO_SPACE(
                "Detect white spaces in fields, please remove it before going any further",
                errors,
            )
